@@ -1,5 +1,5 @@
 #
-# Copyright (C) 2004, 2005 Mekensleep
+# Copyright (C) 2004, 2005, 2006 Mekensleep
 #
 # Mekensleep
 # 24 rue vieille du temple
@@ -25,16 +25,18 @@
 #
 from math import ceil
 from types import StringType
-import time
+from pprint import pformat
+import time, sys
 
 from pokerengine.pokergame import PokerGameServer
+from pokerengine.pokerengineconfig import Config
 
 TOURNAMENT_STATE_ANNOUNCED = "announced"
 TOURNAMENT_STATE_REGISTERING = "registering"
 TOURNAMENT_STATE_RUNNING = "running"
 TOURNAMENT_STATE_COMPLETE = "complete"
             
-def equalizeGames(games, verbose = 0, log_message = None):
+def equalizeCandidates(games):
     #
     # Games less than 70% full are willing to steal players from other
     # games. Games that are more than 70% full and that are not
@@ -50,6 +52,10 @@ def equalizeGames(games, verbose = 0, log_message = None):
         elif not game.isRunning():
             serials = game.serialsAllSorted()
             provide_players.append((game.id, serials[:count - threshold]))
+    return ( want_players, provide_players )
+
+def equalizeGames(games, verbose = 0, log_message = None):
+    ( want_players, provide_players ) = equalizeCandidates(games)
 
     results = []
 
@@ -80,9 +86,8 @@ def equalizeGames(games, verbose = 0, log_message = None):
             if not distributed:
                 break
 
-#    if len(results) > 0:
-#        print "equalizeGames: "
-#        pprint(results)
+    if log_message and verbose > 0 and len(results) > 0:
+        log_message("balanceGames equalizeGames: " + pformat(results))
 
     return results
 
@@ -104,7 +109,7 @@ def breakGames(games, verbose = 0, log_message = None):
         "to_add": [],
         "running": game.isRunning() } for game in games ]
 
-    if verbose > 2: log_message("breakGames: %s" % to_break)
+    if verbose > 2: log_message("balanceGames breakGames: %s" % to_break)
     results = []
     while True:
         result = breakGame(to_break[0], to_break[1:], verbose, log_message)
@@ -115,6 +120,9 @@ def breakGames(games, verbose = 0, log_message = None):
         if len(to_break) < 2:
             break
 
+    if log_message and verbose > 0 and len(results) > 0:
+        log_message("balanceGames breakGames: " + pformat(results))
+
     return results
 
 def breakGame(to_break, to_fill, verbose = 0, log_message = None):
@@ -159,7 +167,6 @@ class PokerTournament:
         self.name = kwargs.get('name', 'no name')
         self.description_short = kwargs.get('description_short', 'nodescription_short')
         self.description_long = kwargs.get('description_long', 'nodescription_long')
-        self.name = kwargs.get('name', 'no name')
         self.serial = kwargs.get('serial', 1)
         self.verbose = kwargs.get('verbose', 0)
         self.players_quota = kwargs.get('players_quota', 10)
@@ -174,26 +181,38 @@ class PokerTournament:
         self.rebuy_delay = kwargs.get('rebuy_delay', 0)
         self.add_on = kwargs.get('add_on', 0)
         self.add_on_delay = kwargs.get('add_on_delay', 60)
-        self.prizes_specs = kwargs.get('prizes_specs', "algorithm")
+        self.prizes_specs = kwargs.get('prizes_specs', "table")
         self.finish_time = -1
         if type(self.start_time) is StringType:
-            self.start_time = int(strftime("%s", strptime(self.start_time, "%Y/%m/%d %H:%M")))
+            self.start_time = int(time.mktime(time.strptime(self.start_time, "%Y/%m/%d %H:%M")))
         self.prefix = ""
         
         self.players = []
+        self.need_balance = False
         self.registered = 0
         self.winners = []
         self.state = TOURNAMENT_STATE_ANNOUNCED
         self.can_register = False
         self.games = []
+        self.id2game = {}
+        
         self.callback_new_state = lambda tournament: True
         self.callback_create_game = lambda tournament: PokerGameServer("poker.%s.xml", tournament.dirs)
         self.callback_game_filled = lambda tournament, game: True
         self.callback_destroy_game = lambda tournament, game: True
         self.callback_move_player = lambda tournament, from_game_id, to_game_id, serial: self.movePlayer(from_game_id, to_game_id, serial)
         self.callback_remove_player = lambda tournament, game_id, serial: self.removePlayer(game_id, serial)
+        self.loadPayouts()
         self.updateRegistering()
 
+    def loadPayouts(self):
+        config = Config(self.dirs)
+        config.load("poker.payouts.xml")
+        self.payouts = []
+        for node in config.header.xpathEval("/payouts/payout"):
+            properties = config.headerNodeProperties(node)
+            self.payouts.append(( int(properties['max']), map(lambda percent: float(percent) / 100, node.content.split())))
+
     def message(self, message):
         print self.prefix + "[PokerTournament %s] " % self.name + message
         
@@ -211,13 +230,13 @@ class PokerTournament:
     def updateRegistering(self):
         if self.state == TOURNAMENT_STATE_ANNOUNCED:
             now = time.time()
-            if now - self.register_time > 0:
+            if now - self.register_time > 0.0:
                 self.changeState(TOURNAMENT_STATE_REGISTERING)
                 return -1
             else:
                 return self.register_time - now
         else:
-            if self.verbose: self.message("updateRegistering: should not be called while tournament is not in announced state")
+            if self.verbose > 0: self.message("updateRegistering: should not be called while tournament is not in announced state")
             return -1
             
     def changeState(self, state):
@@ -231,7 +250,7 @@ class PokerTournament:
         elif self.state == TOURNAMENT_STATE_RUNNING and state == TOURNAMENT_STATE_COMPLETE:
             self.finish_time = time.time()
         else:
-            print "PokerTournament:changeState: cannot change from state %s to state %s" % ( self.state, state )
+            if self.verbose >= 0: print "PokerTournament:changeState: cannot change from state %s to state %s" % ( self.state, state )
             return
         if self.verbose > 2: self.message("state change %s => %s" % ( self.state, state ))
         self.state = state
@@ -288,6 +307,7 @@ class PokerTournament:
         to_game.sit(serial)
         to_game.autoBlindAnte(serial)
         to_player.name = from_player.name
+        to_player.setUserData(from_player.getUserData())
         if(from_player.isSitOut()): to_game.sitOut(serial)
         if(from_player.isBot()): to_game.botPlayer(serial)
         from_game.removePlayer(serial)
@@ -309,11 +329,14 @@ class PokerTournament:
 
             buy_in = game.buyIn()
             for seat in xrange(self.seats_per_game):
+                if not players: break
+                    
                 player = players.pop()
                 game.addPlayer(player)
                 game.payBuyIn(player, buy_in)
                 game.sit(player)
                 game.autoBlindAnte(player)
+                
             self.games.append(game)
             self.callback_game_filled(self, game)
             game.close()
@@ -324,37 +347,35 @@ class PokerTournament:
         loosers = game.serialsBroke()
         loosers_count = len(loosers)
 
-        if loosers_count == 1:
-            self.winners.insert(0, loosers[0])
-            self.callback_remove_player(self, game_id, loosers[0])
-            if self.verbose > 2: self.message("winners %s" % self.winners)
-        elif loosers_count > 1:
-            for serial in loosers:
-                self.winners.insert(0, serial)
-                self.callback_remove_player(self, game_id, serial)
-            if self.verbose > 2: self.message("winners %s" % self.winners)
-
+        for serial in loosers:
+            self.winners.insert(0, serial)
+            self.callback_remove_player(self, game_id, serial)
+        if self.verbose > 2: self.message("winners %s" % self.winners)
+        
         if len(self.winners) + 1 == self.registered:
             game = self.games[0]
             player = game.playersAll()[0]
             self.winners.insert(0, player.serial)
             self.callback_remove_player(self, game.id, player.serial)
-            money = player.money.toint()
-            player.money.set(0)
+            money = player.money
+            player.money = 0
             expected = game.buyIn() * self.registered
-            if money != expected:
+            if money != expected and self.verbose >= 0:
                 self.message("ERROR winner has %d chips and should have %d chips" % ( money, expected ))
-            if self.verbose: self.message("winners %s" % self.winners)
+            if self.verbose > 0: self.message("winners %s" % self.winners)
             self.callback_destroy_game(self, game)
             self.games = []
             self.id2game = {}
             self.changeState(TOURNAMENT_STATE_COMPLETE)
             return False
         else:
-            if loosers_count > 0: self.balanceGames()
+            if loosers_count > 0 or self.need_balance: self.balanceGames()
             return True
         
     def balanceGames(self):
+        self.need_balance = False
+        if len(self.games) < 2: return
+        if self.verbose > 2: self.message("balanceGames")
         to_break = breakGames(self.games, self.verbose, self.message)
         games_broken = {}
         for (from_id, to_id, serials) in to_break:
@@ -369,7 +390,7 @@ class PokerTournament:
                 self.callback_destroy_game(self, game)
                 self.games.remove(game)
                 del self.id2game[game.id]
-            if self.verbose: self.message("balanceGames: broke tables %s" % to_break)
+            if self.verbose > 0: self.message("balanceGames: broke tables %s" % to_break)
             return True
         
         to_equalize = equalizeGames(self.games, self.verbose, self.message)
@@ -377,6 +398,10 @@ class PokerTournament:
             if self.verbose > 2: self.message("balanceGames: player %d moved from %d to %d" % ( serial, from_id, to_id ))
             self.callback_move_player(self, from_id, to_id, serial)
 
+        ( want_players, provide_players ) = equalizeCandidates(self.games)
+        self.need_balance = want_players and not provide_players
+        if self.need_balance and self.verbose > 2: self.message("balanceGames: postponed game equalization")
+        
         return len(to_equalize) > 0
 
     def prizes(self, buy_in):
@@ -384,6 +409,8 @@ class PokerTournament:
             return None
         if self.prizes_specs == "algorithm":
             return self.prizesAlgorithm(buy_in)
+        elif self.prizes_specs == "table":
+            return self.prizesTable(buy_in)
         else:
             return None
 
@@ -420,3 +447,15 @@ class PokerTournament:
         prizes[0] += rest
         return prizes
                 
+    def prizesTable(self, buy_in):
+        for ( max, payouts ) in self.payouts:
+            if self.registered <= max:
+                break
+
+        total = self.registered * buy_in
+        prizes = map(lambda percent: int(total * percent), payouts)
+        #
+        # What's left because of rounding errors goes to the tournament winner
+        #
+        prizes[0] += total - sum(prizes)
+        return prizes
